@@ -1,13 +1,45 @@
 from .utils import to_interval, timeit, prod, rrange
 import os
 import xarray as xr
-import numpy as np  # type: ignore
+import numpy as np
 import re
 import typing
 
+try:
+    from numpy.typing import DTypeLike
+except ImportError:
+    # Workaround for python 3.6
+    DTypeLike = typing.Type  # type: ignore
 
-def _fromfile(f: typing.TextIO, *, count: int, dtype: np.dtype, **kwargs) -> np.ndarray:
-    ret = np.empty(count, dtype=dtype)
+
+def _fromfile(
+    f: typing.TextIO, *, count: int, dtype: DTypeLike, **kwargs
+) -> np.ndarray:
+    """
+    Read count amount of dtype from the textio stream, i.e. a file
+    opened for reading. It always reads full lines. If there is more
+    data on the line then it should read, all data is returened. If
+    less data is read, then a shorter array is returned. Thus the
+    calling function needs to handle the case that more or less data
+    is returned.
+
+    Parameters
+    ----------
+    f : textio
+        the opened file
+    count : int
+        amount that should be read
+    dtype : dtype
+        the datatype to expect
+    kwargs : dict
+        Additional arguments for np.fromstring
+
+    Returns
+    -------
+    np.ndarray
+        The read data.
+    """
+    ret = np.empty(count, dtype=dtype)  # type: np.ndarray
     pos = 0
     bad = re.compile(r"(\d)([+-]\d)")
 
@@ -31,7 +63,8 @@ def _fromfile(f: typing.TextIO, *, count: int, dtype: np.dtype, **kwargs) -> np.
         if ln == 0 or count <= pos:
             break
     if count != pos:
-        return ret[:pos]
+        ret = ret[:pos]
+        return ret
     return ret
 
 
@@ -88,16 +121,22 @@ def read_magnetic_field(fn: str, ds: xr.Dataset) -> xr.DataArray:
     xr.Dataset
         The magenetic field strength
     """
-    shape = ds.R_bounds.shape
-    assert len(shape) == 6
-    shape = [i + 1 for i in shape[:3]]
+    if "R_bounds" in ds:
+        shape = ds.R_bounds.shape
+        assert len(shape) == 6
+        shape = [i + 1 for i in shape[:3]]
+        dims = ds.R_bounds.dims[:3]
+    else:
+        shape = ds._plasma_map.shape
+        shape = [i + 1 for i in shape]
+        dims = ds._plasma_map.dims
     nx, ny, nz = shape
     with open(fn) as f:
         raw = _fromfile(f, dtype=float, count=nx * ny * nz, sep=" ")
         _assert_eof(f, fn)
     raw = raw.reshape(shape[::-1])
     raw = np.swapaxes(raw, 0, 2)
-    return to_interval(ds.R_bounds.dims[:3], raw)
+    return to_interval(dims, raw)
 
 
 def write_magnetic_field(path: str, ds: xr.Dataset) -> None:
@@ -274,9 +313,24 @@ def write_mappings(da: xr.DataArray, fn: str) -> None:
 
 
 def get_locations(path: str, ds: xr.Dataset = None) -> xr.Dataset:
+    """
+    Read locations from folder path and add to dataset (if given).
+
+    Parameters
+    ----------
+    path : str
+        folder from which to read
+    ds : xr.Dataset (optional)
+        The dataset to add to. If not given return a new dataset.
+
+    Returns
+    -------
+    xr.Dataset
+        A dataset with the coordinates set
+    """
     if ds is None:
         ds = xr.Dataset()
-    ds: xr.Dataset
+    assert isinstance(ds, xr.Dataset)
     phi, r, z = read_locations(path + "/GRID_3D_DATA")
     ds = ds.assign_coords(
         {
@@ -287,27 +341,26 @@ def get_locations(path: str, ds: xr.Dataset = None) -> xr.Dataset:
     )
     ds.emc3.unit("R_bounds", "m")
     ds.emc3.unit("z_bounds", "m")
-    return ds
-
-
-def get_locations_old(path: str):
-    phi, r, z = read_locations(path + "/GRID_3D_DATA")
-
-    ds = xr.Dataset()
-    ds = ds.assign_coords({"R": (("r", "theta", "phi"), r)})
-    ds = ds.assign_coords({"z": (("r", "theta", "phi"), z)})
-    ds = ds.assign_coords({"phi": phi})
-    for d in "Rz":
-        ds[d].attrs["units"] = "meters"
+    assert isinstance(ds, xr.Dataset)
     return ds
 
 
 def scrape(f: typing.TextIO, *, ignore=None, verbose=False) -> str:
-    """read next data line from configuration file (skip lines with leading *)
-    f : fileobject to read from
-    ignore : (str or None) if any of the characters are given,
-             interpret as comment and remove. In this case only return
-             non-empty lines.
+    """read next data line from configuration file (skip lines with
+    leading *)
+
+    Parameters
+    ----------
+    f : textio
+        fileobject (opened file) to read from
+    ignore : (str or None)
+        if any of the characters are given, interpret as comment and
+        remove. In this case only return non-empty lines.
+
+    Returns
+    ------
+    str
+        the first non-ignored line
     """
     while True:
         s = f.readline()
@@ -326,19 +379,49 @@ def scrape(f: typing.TextIO, *, ignore=None, verbose=False) -> str:
 
 
 def _assert_eof(f: typing.TextIO, fn: str) -> None:
+    """Ensure the opened file is read to the end.
+
+    Parameters
+    ----------
+    f : textio
+        the opened file
+    fn : str
+        The name of the file, used for the error.
+
+    Raises
+    ------
+    AssertionError
+        if the file is not at the end.
+    """
     test = _fromfile(f, dtype=float, count=2, sep=" ")
-    if len(test):
-        raise RuntimeError(f"Expected EOF, but found more data in {fn}")
+    assert len(test) == 0, f"Expected EOF, but found more data in {fn}"
 
 
-def read_plate(cwd: str, fn: str = "") -> typing.Tuple[np.ndarray, ...]:
-    with open(cwd + fn) as f:
+def read_plate(filename: str) -> typing.Tuple[np.ndarray, ...]:
+    """
+    Read Target structures from a file that is in the Kisslinger
+    format as used by EMC3. It returns the coordinates as plain array
+    in the order or R, z, phi.
+
+    Parameters
+    ----------
+    filename : str
+        The location of the file to read
+
+    Returns
+    -------
+    np.ndarray
+        The major radius coordinates of the corners in meters.
+    np.ndarray
+        The z coordinate of the corners in meters
+    np.ndarray
+        the phi coordinate of the corners in radian
+    """
+    with open(filename) as f:
         # first line is a comment ...
         _ = next(f)
         setup = next(f).split()
-        if len(setup) != 5:
-            print(setup)
-            raise RuntimeError("Expected 5 values!")
+        assert len(setup) == 5, f"Expected 5 values but got {setup}"
         for zero in setup[3:]:
             assert float(zero) == 0.0
         nx, ny, nz = [int(i) for i in setup[:3]]
@@ -348,16 +431,16 @@ def read_plate(cwd: str, fn: str = "") -> typing.Tuple[np.ndarray, ...]:
         for x in range(nx):
             phi[x] = float(next(f)) / 180 * np.pi
             for y in range(ny):
+                s = next(f)
                 try:
                     try:
-                        s = next(f)
                         r[x, y], z[x, y] = [float(i) / 100 for i in s.split()]
                     except ValueError:
                         s = s.split("!")[0]
                         r[x, y], z[x, y] = [float(i) / 100 for i in s.split()]
                 except ValueError:
-                    raise ValueError(f"Error with {s} in {cwd+fn}")
-        _assert_eof(f, cwd + fn)
+                    raise ValueError(f"Error with {s} in {filename}")
+        _assert_eof(f, filename)
         return (r, z, phi)
 
 
@@ -365,18 +448,30 @@ plates_labels = ["f_n", "f_E", "avg_n", "avg_Te", "avg_Ti"]
 
 
 def read_plates_raw(cwd: str, fn: str) -> typing.Sequence[xr.Dataset]:
+    """
+    Read the target mapped info to a list of xr.Datasets
+
+    Parameters
+    ----------
+    cwd : str
+        The folder in which to look for the files (with trailing slash)
+    fn : str
+        the relative name of the file to read
+
+    Returns
+    -------
+    list of xr.Dataset
+        The read data as a list of datasets
+    """
     with open(cwd + fn) as f:
         s = scrape(f, ignore="!").split()
         _, num_plates = [int(i) for i in s]
         plates = []
         for plate in range(num_plates):
             s = scrape(f, ignore="!").split()
-            try:
-                _, geom = s
-            except:
-                print(plate, ":", s)
-                raise
-            r, z, phi = read_plate(cwd, geom)
+            assert len(s) == 2, f"{plate} : {s}"
+            _, geom = s
+            r, z, phi = read_plate(cwd + geom)
             nx, ny = r.shape
             nx -= 1
             ny -= 1
@@ -520,6 +615,9 @@ def read_plates_raw(cwd: str, fn: str) -> typing.Sequence[xr.Dataset]:
 
 
 def plates_raw_to_ds(plates: typing.Sequence[xr.Dataset]) -> xr.Dataset:
+    """
+    Convert a list of datasets to one dataset
+    """
     dims = {d: 0 for d in plates[0].dims}
     for plate in plates:
         for k, v in plate.dims.items():
@@ -561,6 +659,9 @@ def plates_raw_to_ds(plates: typing.Sequence[xr.Dataset]) -> xr.Dataset:
 
 
 def load_plates(cwd: str) -> xr.Dataset:
+    # Deprecate?
+    if cwd[-1] != "/":
+        cwd += "/"
     with timeit("\nReading raw: %f"):
         plates = read_plates_raw(cwd, "TARGET_PROFILES")
     with timeit("To xarray: %f"):
@@ -568,6 +669,7 @@ def load_plates(cwd: str) -> xr.Dataset:
 
 
 def write_plates(cwd: str, plates: xr.Dataset) -> None:
+    # Deprecate?
     with timeit("Writing ncs: %f"):
         # Note we really should compress to get rid of the NaN's we added
         plates.to_netcdf(
@@ -580,8 +682,10 @@ def write_plates(cwd: str, plates: xr.Dataset) -> None:
 
 
 def read_plates(cwd: str) -> xr.Dataset:
-    # with timeit("Reading ncs: %f"):
-    return xr.open_dataset(f"{cwd}/TARGET_PROFILES.nc")
+    # Deprecate?
+    ds = xr.open_dataset(f"{cwd}/TARGET_PROFILES.nc")
+    assert isinstance(ds, xr.Dataset)
+    return ds
 
 
 def get_plates(cwd: str, cache: bool = True) -> xr.Dataset:
@@ -615,8 +719,60 @@ def get_plates(cwd: str, cache: bool = True) -> xr.Dataset:
     return load_plates(cwd)
 
 
-def read_mapped_nice(
-    dir: str, var: str, mapping: typing.Union[None, xr.Dataset, xr.DataArray] = None
+def ensure_mapping(
+    dir: str,
+    mapping: typing.Union[None, xr.Dataset, xr.DataArray] = None,
+    need_mapping: bool = True,
+    fn: typing.Union[None, str] = None,
+) -> xr.Dataset:
+    """
+    Ensure that basic infos are present to read datafiles.
+
+    Parameters
+    ----------
+    dir : str
+        The folder in which to look for files
+    mapping : None or xr.Dataset or xr.DataArray
+        Either the required mapping, in which case nothing is done, or
+        None, in which case the data is read.
+    need_mapping : bool (optional)
+        If true the mapping infomation has to be loaded, otherwise the
+        shape of the data is sufficient. Defaults to True.
+    fn : str (optional)
+        The file to read, used for a potential error message
+
+    Returns
+    -------
+    xr.Dataset
+        A dataset with the required information
+
+    Raises
+    ------
+    FileNotFoundError
+        If the required info could not be read
+    """
+    if dir == "":
+        dir = "."
+    if mapping is None:
+        try:
+            mapping = get_locations(dir)
+            if need_mapping:
+                mapping = read_fort_file(mapping, f"{dir}/fort.70", **files["fort.70"])
+        except FileNotFoundError:
+            raise FileNotFoundError(
+                f"""Reading {fn+ ' ' if fn is not None else ''} mapped requires mapping information, but the required
+infomation in '{dir}' could not be found.  Ensure all files are present
+in the folder or pass in a dataset that contains the mapping
+informations."""
+            )
+    else:
+        if isinstance(mapping, xr.DataArray):
+            mapping = xr.Dataset(dict(_plasma_map=mapping))
+    return mapping
+
+
+def read_var(
+    dir: str, var: str, ds: typing.Union[None, xr.Dataset, xr.DataArray] = None
 ) -> xr.Dataset:
     """
     Read the variable var from the simulation in directory dir.
@@ -627,22 +783,16 @@ def read_mapped_nice(
         The path of the simulation directory
     var: str
         The name of the variable to be read
-    mapping: xr.DataArray or xr.Dataset or None (optional)
-        If present the mapping. If not present, the information is
-        read from the simulation directory.
+    ds: xr.DataArray or xr.Dataset or None (optional)
+        An xemc3 Dataset or an DataArray containing the mapping. If
+        not given and a mapping is required to read the file, it is
+        read from the directory.
 
     Returns
     -------
     xr.Dataset
         A dataset in wich at least ``var`` is set.
     """
-    if mapping is None:
-        mapping = get_locations(dir)
-        mapping = read_fort_file(mapping, f"{dir}/fort.70", **files["fort.70"])
-    else:
-        if isinstance(mapping, xr.DataArray):
-            mapping = xr.Dataset(dict(_plasma_map=mapping))
-
     for fn, fd in files.items():
         if "vars" not in fd:
             continue
@@ -652,7 +802,9 @@ def read_mapped_nice(
                 and var.startswith(v[:-2])
                 and var[len(v[:-2]) :].isdigit()
             ):
-                return read_fort_file(mapping, f"{dir}/{fn}", **fd)
+                ds = ensure_mapping(dir, ds, fd.get("type", "mapped") == "mapped")
+                assert isinstance(ds, xr.Dataset)
+                return read_fort_file(ds, f"{dir}/{fn}", **fd)
     raise ValueError(f"Don't know how to read {var}")
 
 
@@ -662,33 +814,38 @@ def read_mapped(
     skip_first: int = 0,
     ignore_broken: bool = False,
     kinetic: bool = False,
-    dtype: np.dtype = float,
+    dtype: DTypeLike = float,
+    squeeze: bool = True,
 ) -> typing.Sequence[xr.DataArray]:
     """
     Read a file with the emc3 mapping.
 
     Parameters
     ----------
-    fn: str
+    fn : str
         The full path of the file to read
-    mapping: xr.DataArray or xr.Dataset
+    mapping : xr.DataArray or xr.Dataset
         The dataarray of the mesh mapping or a dataset containing the
         mapping
-    skip_first: int (optional)
+    skip_first : int (optional)
         Ignore the first n lines. Default 0
     ignore_broken: boolean (optional)
         if incomplete datasets at the end should be ignored. Default: False
-    kinetic: boolen (optional)
+    kinetic : boolen (optional)
         The file contains also data for cells that are only evolved by
         EIRENE, rather then EMC3. Default: False
-    dtype: datatype (optional)
+    dtype : datatype (optional)
         The type of the data, e.g. float or int. Is passed to
         numpy. Default: float
+    squeeze : bool
+        If True return a DataArray if only one field is read.
 
     Returns
     -------
-    list of xr.DataArray
-        The data that has been read from the file
+    xr.DataArray or list of xr.DataArray
+        The data that has been read from the file. If squeeze is True
+        and only one field is read only a sinlge DataArray is
+        returend.
     """
 
     if isinstance(mapping, xr.Dataset):
@@ -736,6 +893,8 @@ def read_mapped(
     if skip_first:
         for first, da in zip(firsts, das):
             da.attrs["print_before"] = first
+    if squeeze and len(das) == 1:
+        das = das[0]
     return das
 
 
@@ -782,10 +941,9 @@ def write_mapped_nice(ds: xr.Dataset, dir: str, fn: str = None, **args) -> None:
                 at = datas[i].attrs
                 datas[i] = datas[i] / ops["scale"]
                 datas[i].attrs = at
-        if datas == []:
-            raise AssertionError(
-                f"Requested to write file {dir}/{fn} but required data not found."
-            )
+        assert (
+            datas != []
+        ), f"Requested to write file {dir}/{fn} but required data not found."
         write_mapped(datas, ds["_plasma_map"], f"{dir}/{fn}", **meta)
 
 
@@ -1091,20 +1249,55 @@ else:
     _files_bak = files
 
 
-def read_fort_file(ds, fn, type="mapped", **opts):
+def read_fort_file_pub(
+    fn: str, ds: typing.Union[None, xr.Dataset, xr.DataArray] = None, **opts
+) -> xr.Dataset:
+    """
+    Read a EMC3 simulation file. The expected content is derived from
+    the filename.
+
+    Parameters
+    ----------
+    fn : str
+        The location of the file to read
+    ds : xr.DataArray or xr.Dataset or None
+        The mapping or a dataset containing a mapping or None. If one
+        is needed and none is provided, it is read from the folder.
+    opts : dict
+        Additional options depending on the type of file that is read.
+
+    Returns
+    -------
+    xr.Dataset
+        The read variable is set. If a Dataset was given, the newly
+        read entries from the file are added. Otherwise a new Dataset
+        is returned with the read variables added.
+    """
+    filename = fn.split("/")[-1]
+    defaults = files[filename].copy()
+    defaults.update(opts)
+    type = defaults.get("type", "mapped")
+    ds = ensure_mapping("/".join(fn.split("/")[:-1]), ds, type == "mapped")
+    assert isinstance(ds, xr.Dataset)
+    return read_fort_file(ds, fn, **defaults)
+
+
+def read_fort_file(ds: xr.Dataset, fn: str, type: str = "mapped", **opts) -> xr.Dataset:
+    """
+    Read an EMC3 simulation file and add to a dataset.
+
+    """
     assert files == _files_bak
     if type == "mapping":
         opts.pop("vars", False)
         ds["_plasma_map"] = read_mappings(fn, ds.R_bounds.data.shape[:3])
-        if opts != {}:
-            raise RuntimeError(
-                "Unexpected arguments: "
-                + ", ".join([f"{k}={v}" for k, v in opts.items()])
-            )
+        assert opts == {}, "Unexpected arguments: " + ", ".join(
+            [f"{k}={v}" for k, v in opts.items()]
+        )
         return ds
     elif type == "mapped":
         vars = opts.pop("vars")
-        datas = read_mapped(fn, ds["_plasma_map"], **opts)
+        datas = read_mapped(fn, ds["_plasma_map"], **opts, squeeze=False)
         opts = {}
     elif type == "full":
         vars = opts.pop("vars")
@@ -1115,10 +1308,9 @@ def read_fort_file(ds, fn, type="mapped", **opts):
     else:
         raise RuntimeError(f"Unexpected type {type}")
     assert files == _files_bak
-    if opts != {}:
-        raise RuntimeError(
-            "Unexpected arguments: " + ", ".join([f"{k}={v}" for k, v in opts.items()])
-        )
+    assert opts == {}, "Unexpected arguments: " + ", ".join(
+        [f"{k}={v}" for k, v in opts.items()]
+    )
     vars = vars.copy()
     assert files == _files_bak
     assert opts == {}
@@ -1130,10 +1322,9 @@ def read_fort_file(ds, fn, type="mapped", **opts):
             vars[key % i] = flexi
 
     assert files == _files_bak
-    if len(vars) != len(datas):
-        raise RuntimeError(
-            f"in file {fn} we found {len(datas)} fields but only {len(vars)} vars are given!"
-        )
+    assert len(vars) == len(
+        datas
+    ), f"in file {fn} we found {len(datas)} fields but only {len(vars)} vars are given!"
     assert files == _files_bak
     for (var, varopts), data in zip(vars.items(), datas):
         ds[var] = data
